//! Subscription manager
use crate::{client, Error};
use futures::future::join_all;
use nostr_rs_subscription_manager::{
    ActiveSubscription as ActiveSubscriptionInner, SubscriptionManager,
};
use nostr_rs_types::{
    client::subscribe::{self, is_all_events},
    relayer,
    types::SubscriptionId,
    Response,
};
use std::sync::{
    atomic::{AtomicUsize, Ordering},
    Arc,
};
use tokio::sync::{mpsc, Mutex};
use url::Url;

use super::AllClients;

#[derive(Debug, Copy, Default, Eq, PartialEq, Clone)]
/// Subscription status
pub enum Status {
    /// Subscription is awaiting to be subscribed
    #[default]
    Queued,
    /// Subscribed is active
    Subscribed,
    /// Technically unsubscribed, and fetching future events from the relayer
    /// from the All-Events meta subscription
    Stale,
}

#[derive(Debug, Default)]
struct SubscriptionInner {
    /// Active subscription (in the client side), when this is Drop all clients unsubscribes
    active_subscription: Option<Vec<client::ActiveSubscription>>,
    /// Subscription status
    status: Arc<Mutex<Status>>,
    /// raw request
    subscription_request: subscribe::Subscribe,
}

/// Active subscription
///
/// This object is responsible for keeping track of a subscription for a
/// connection
///
/// This must be dropped to unsubscribe from the subscription manager
pub struct ActiveSubscription {
    _id: ActiveSubscriptionInner<PoolSubscriptionId, SubscriptionInner>,
    status: Arc<Mutex<Status>>,
    active_subscriptions: Arc<AtomicUsize>,
    queued_subscriptions: Arc<AtomicUsize>,
    stale_subscriptions: Arc<AtomicUsize>,
}

impl Drop for ActiveSubscription {
    fn drop(&mut self) {
        let active_subscriptions = self.active_subscriptions.clone();
        let queued_subscriptions = self.queued_subscriptions.clone();
        let stale_subscriptions = self.stale_subscriptions.clone();
        let status = self.status.clone();

        tokio::spawn(async move {
            match *status.lock().await {
                Status::Subscribed => active_subscriptions.fetch_sub(1, Ordering::Relaxed),
                Status::Queued => queued_subscriptions.fetch_sub(1, Ordering::Relaxed),
                Status::Stale => stale_subscriptions.fetch_sub(1, Ordering::Relaxed),
            }
        });
    }
}

/// Pool subscription ID
#[derive(Debug, Clone, Eq, PartialEq, Hash, Ord, PartialOrd)]
pub struct PoolSubscriptionId((SubscriptionId, Option<Url>));

impl Default for PoolSubscriptionId {
    fn default() -> Self {
        Self((SubscriptionId::empty(), None))
    }
}

/// Subscription manager
///
/// Clients who are added to the pool are automatically subscribed to all
/// events, which is known as the All-Events subscription.
///
/// Subscriptions in the client pool are smarter than the raw subscriptions at
/// the client level. These subscriptions will be active until an "End of
/// stored" event is received; the client pool will unsubscribe at that point,
/// and the All-Events subscriptions, filtered internally, will fetch future
/// events. By doing so, only past events are queried, and there is a single
/// stream of future events to be a good citizen with other relayers.
#[derive(Default)]
pub(crate) struct Manager {
    subscription_manager: Arc<SubscriptionManager<PoolSubscriptionId, SubscriptionInner>>,
    all_clients: AllClients,
    active_subscriptions: Arc<AtomicUsize>,
    queued_subscriptions: Arc<AtomicUsize>,
    stale_subscriptions: Arc<AtomicUsize>,
}

/// Maximum number of subscriptions
pub const MAX_SUBSCRIPTIONS: usize = 50;

impl Manager {
    /// Processes all messages from the client pools
    ///
    /// The client pool creates a subscription to the All-Events subscription,
    /// this callback checks if there are any listener to this event, otherwise
    /// it will not process the event.
    pub async fn process_message(
        self: &Arc<Self>,
        message: Response,
        url: Url,
        return_to: mpsc::Sender<(Response, Url)>,
    ) -> Result<(), Error> {
        match message {
            Response::EndOfStoredEvents(subscription_id) => {
                let subscription_id = PoolSubscriptionId((subscription_id.0, None));
                let mut subscription = self.subscription_manager.subbscriptions_mut().await;
                if let Some(s) = subscription.get_mut(&subscription_id) {
                    *s.status.lock().await = Status::Stale;
                    let _ = s.active_subscription.take();

                    self.active_subscriptions.fetch_sub(1, Ordering::Relaxed);
                    self.stale_subscriptions.fetch_add(1, Ordering::Relaxed);
                }

                return_to
                    .try_send((
                        Response::EndOfStoredEvents(subscription_id.0 .0.into()),
                        url,
                    ))
                    .map_err(|e| Error::InternalChannel(e.to_string()))?;

                Ok(())
            }
            Response::Event(relayer::Event {
                subscription_id,
                event,
            }) => {
                if !is_all_events(&subscription_id) {
                    // This is not an All-Events subscription, it must be passed on as it is an stored event
                    return_to
                        .try_send((
                            Response::Event(relayer::Event {
                                subscription_id,
                                event,
                            }),
                            url.clone(),
                        ))
                        .map_err(|e| Error::InternalChannel(e.to_string()))?;
                    return Ok(());
                }

                for id in self.subscription_manager.get_subscribers(&event).await {
                    return_to
                        .try_send((
                            Response::Event(relayer::Event {
                                subscription_id: id.0 .0.clone(),
                                event: event.clone(),
                            }),
                            url.clone(),
                        ))
                        .map_err(|e| Error::InternalChannel(e.to_string()))?;
                }

                Ok(())
            }
            any_message => {
                return_to
                    .try_send((any_message, url))
                    .map_err(|e| Error::InternalChannel(e.to_string()))?;
                Ok(())
            }
        }
    }

    async fn update_active_subscriptions(self: &Arc<Self>) {
        if self.active_subscriptions.load(Ordering::Relaxed) >= MAX_SUBSCRIPTIONS
            || self.queued_subscriptions.load(Ordering::Relaxed) == 0
        {
            return;
        }

        let clients = self.all_clients.read().await;
        let mut subscriptions = self.subscription_manager.subbscriptions_mut().await;

        for subscription in subscriptions.values_mut() {
            if *subscription.status.lock().await == Status::Queued {
                let wait_all = clients
                    .values()
                    .map(|(_, (_, sender))| {
                        sender.subscribe(subscription.subscription_request.clone())
                    })
                    .collect::<Vec<_>>();

                if let Ok(active_subscriptions) = join_all(wait_all)
                    .await
                    .into_iter()
                    .collect::<Result<Vec<_>, _>>()
                {
                    subscription.active_subscription = Some(active_subscriptions);
                    *subscription.status.lock().await = Status::Subscribed;

                    let queued_subscribed =
                        self.queued_subscriptions.fetch_sub(1, Ordering::Relaxed);
                    let active_subscriptions =
                        self.active_subscriptions.fetch_add(1, Ordering::Relaxed);
                    if queued_subscribed == 0 || active_subscriptions >= MAX_SUBSCRIPTIONS {
                        break;
                    }
                }
            }
        }
    }

    /// Creates a new subscription with a given filters
    pub async fn subcribe(
        self: &Arc<Self>,
        subscription_request: subscribe::Subscribe,
        specific_url: Option<Url>,
    ) -> ActiveSubscription {
        let status = Arc::new(Mutex::new(Status::Queued));
        let id = self
            .subscription_manager
            .subscribe(
                PoolSubscriptionId((
                    subscription_request.subscription_id.clone(),
                    specific_url.clone(),
                )),
                subscription_request.filters.clone(),
                SubscriptionInner {
                    status: status.clone(),
                    active_subscription: None,
                    subscription_request,
                },
            )
            .await;

        self.queued_subscriptions.fetch_add(1, Ordering::Relaxed);

        let this = self.clone();
        tokio::spawn(async move {
            this.update_active_subscriptions().await;
        });

        ActiveSubscription {
            _id: id,
            status,
            active_subscriptions: self.active_subscriptions.clone(),
            queued_subscriptions: self.queued_subscriptions.clone(),
            stale_subscriptions: self.stale_subscriptions.clone(),
        }
    }

    /// Total number of subscribers
    pub fn total_subscribers(&self) -> usize {
        self.active_subscriptions.load(Ordering::Relaxed)
            + self.queued_subscriptions.load(Ordering::Relaxed)
            + self.stale_subscriptions.load(Ordering::Relaxed)
    }
}
