//! Local connection
//!
//! Add types for adding local connections
use crate::{connection::ConnectionId, Error, Relayer};
use nostr_rs_storage_base::Storage;
use nostr_rs_types::{Request, Response};
use std::sync::Arc;
use tokio::sync::mpsc::{Receiver, Sender};

/// Local connection
pub struct LocalConnection<T>
where
    T: Storage + Send + Sync + 'static,
{
    sender: Sender<(ConnectionId, Request)>,
    receiver: Receiver<Response>,
    pub(crate) conn_id: ConnectionId,
    relayer: Arc<Relayer<T>>,
}

impl<T> LocalConnection<T>
where
    T: Storage + Send + Sync + 'static,
{
    /// Receive a message from the relayer
    pub async fn recv(&mut self) -> Option<Response> {
        self.receiver.recv().await
    }

    /// Try to receive a message from the relayer
    pub fn try_recv(&mut self) -> Option<Response> {
        self.receiver.try_recv().ok()
    }

    /// Sends a request to the relayer
    pub async fn send(&self, request: Request) -> Result<(), Error> {
        self.sender
            .send((self.conn_id, request))
            .await
            .map_err(|e| Error::LocalSendError(Box::new(e)))
    }
}

impl<T> Drop for LocalConnection<T>
where
    T: Storage + Send + Sync + 'static,
{
    fn drop(&mut self) {
        self.relayer.drop_connection(self);
    }
}

impl<T>
    From<(
        ConnectionId,
        Receiver<Response>,
        Sender<(ConnectionId, Request)>,
        Arc<Relayer<T>>,
    )> for LocalConnection<T>
where
    T: Storage + Send + Sync + 'static,
{
    fn from(
        value: (
            ConnectionId,
            Receiver<Response>,
            Sender<(ConnectionId, Request)>,
            Arc<Relayer<T>>,
        ),
    ) -> Self {
        LocalConnection {
            conn_id: value.0,
            receiver: value.1,
            sender: value.2,
            relayer: value.3,
        }
    }
}
